#   Copyright 2018ff. Stephan Druskat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from resosuma.graph.node import Node
from resosuma.graph.edge import Edge
import csv
from resosuma.exceptions import CSVInputError


class ActivityGraph:
    '''
    A directed graph representing activities in the research software
    sustainability space.
    '''

    def __init__(self):
        self.nodes = set()
        self.edges = set()

    def add_activity(self, source: str, activity: str, target: str):
        '''
        Add an activity to the activity graph by mapping the provided
        string arguments to nodes and edges

        Keyword arguments:
        source -- the name of the source node of the activity
        activity -- the name of the activity
        target -- the name of the target node of the activity
        '''
        edge = Edge(source, activity, target)
        self.nodes.add(Node(source))
        self.nodes.add(Node(target))
        self.edges.add(edge)
        return edge

    def get_activities(self):
        return self.edges

    def get_nodes(self):
        return self.nodes

    def read_csv(self, file_path: str):
        '''
        Read a CSV file and map it to the activity graph.

        The CSV file must have three columns, of which the first
        represents the actor of an action, the second the action itself,
        and the third the actee of the action.

        Although the function tries to sniff the CSV dialect, it is
        suggested that the CSV file uses the default delimiter (,) and
        quotechar (").

        Keyword arguments:
        file_path -- the path to the CSV file to be read
        '''
        with open(file_path) as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(1024))
            csvfile.seek(0)
            reader = csv.reader(csvfile, dialect)
            for idx, row in enumerate(reader):
                if len(row) != 3:
                    raise(CSVInputError(str(row),
                                        "CSV row has more or less " +
                                        "than 3 columns " +
                                        "(line " + str(idx + 1) + ")."))
                self.add_activity(row[0], row[1], row[2])



#   Copyright 2018ff. Stephan Druskat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from resosuma.graph.activitygraph import ActivityGraph as AG
from graphviz import Digraph


def create(csvpath: str, outputpath: str, format: str):
    '''
    Creates an activity graph from an input CSV file and visualizes the
    graph via graphviz.

    The CSV file must have 3 columns, of which the first contains
    actors of an action, the second the action, and the third the actee
    of the action.the

    Keyword arguments:
    csvpath --  the path of the CVS file
    outputpath -- the path of the outputfile (also the filename of the)
                  graphviz data file
    format -- the format of the visualization output (svg, pdf, png)
    '''
    graph = AG()
    graph.read_csv(csvpath)
    dot = Digraph(comment='Research Software Sustainability Space Graph',
                  format=format)
    dot.attr(rankdir='LR', size='8,5')
    for n in graph.get_nodes():
        dot.node(n.get_label(), n.get_label())
    for e in graph.get_activities():
        # Edges with a label "are" are rendered differently
        if e.get_label() != "are":
            if e.get_target() != e.get_source():
                # If source and target aren't the same, render as per
                # default, i.e., as a gray edge.
                dot.edge(e.get_source().get_label(),
                         e.get_target().get_label(), e.get_label(),
                         color='gray66')
    for e in graph.get_activities():
        # Edges with a label "are" are rendered differently
        if e.get_label() != "are":
            if e.get_target() == e.get_source():
                # If source and target are the same, render so that the
                # edge and label are more readable, i.e., as a black edge.
                dot.edge(e.get_source().get_label(),
                         e.get_target().get_label(), e.get_label(),
                         color='black', rank='max')
        else:
            dot.edge(e.get_source().get_label(),
                     e.get_target().get_label(), e.get_label(),
                     color='firebrick1', rank='max')
    # Render and show in the default viewer
    dot.render(outputpath, view=True)



#   Copyright 2018ff. Stephan Druskat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from resosuma.graph.node import Node


class Edge:
    '''
    An edge in a directed graph. Edges have a source and a target node.

    Attributes:
    source_node -- A resosuma.graph.node.Node, the source of the edge
    target_node -- A resosuma.graph.node.Node, the target of the edge
    label -- The label string of the edge
    '''

    def __init__(self, source: str, label: str, target: str):
        '''
        Constructor setting attributes.

        Note that this function explicitly creates Node objects from
        the source and target arguments.

        Keyword arguments:
        source -- the name of the source node
        label -- the name of the edge
        target -- the name of the target node
        '''
        self.source_node = Node(source)
        self.label = label
        self.target_node = Node(target)

    def get_source(self):
        return self.source_node

    def get_target(self):
        return self.target_node

    def get_label(self):
        return self.label

    def __str__(self):
        src = self.source_node.get_label()
        trg = self.target_node.get_label()
        return 'Edge: ' + src + ' > ' + self.label + ' > ' + trg

    def __key(self):
        return (self.source_node, self.label, self.target_node)

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __hash__(self):
        return hash(self.__key())



#   Copyright 2018ff. Stephan Druskat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


class Node:
    '''
    A node in a directed graph.
    '''

    def __init__(self, label: str):
        self.label = label

    def get_label(self):
        '''
        Returns the label of the node.
        '''
        return self.label

    def __key(self):
        return self.label

    def __str__(self):
        return 'Node: ' + self.get_label()

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __hash__(self):
        return hash(self.__key())



#   Copyright 2018ff. Stephan Druskat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import namedtuple

__version__ = '0.2.0'

version_info = namedtuple('version_info',
                          ['major', 'minor', 'micro', 'releaselevel'])

major_minor_micro = __version__.split('-')[0]

s = major_minor_micro.split('.')

version_info.major = int(s[0])
version_info.minor = int(s[1])
version_info.micro = int(s[2])

version_info.releaselevel = __version__[len(major_minor_micro) + 1:]



from setuptools import setup
import os
import sys

_here = os.path.abspath(os.path.dirname(__file__))

if sys.version_info[0] < 3:
    with open(os.path.join(_here, 'README.rst')) as f:
        long_description = f.read()
else:
    with open(os.path.join(_here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()

version = {}
with open(os.path.join(_here, 'resosuma', 'version.py')) as f:
    exec(f.read(), version)

setup(
    name='resosuma',
    version=version['__version__'],
    description=('A package providing mappings of the '
                 'research software sustainability space.'),
    long_description=long_description,
    author='Stephan Druskat',
    author_email='mail@sdruskat.net',
    url='https://github.com/research-software/resosuma',
    license='Apache-2.0',
    packages=['resosuma', 'resosuma.graph'],
    install_requires=[
        'graphviz==0.8.3'
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.6']
)
